// Package response provides wrapper gin returns json data in the same format.
package response

import (
	"encoding/json"
	"fmt"
	"net/http"
	"time"

	"github.com/gin-gonic/gin"

	"gitlab.wanzhuangkj.com/tush/xpkg/errcode"
	"gitlab.wanzhuangkj.com/tush/xpkg/gin/xctx"
	ctxUtil "gitlab.wanzhuangkj.com/tush/xpkg/gin/xctx"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"
)

const CustomErrorCode = 0

// Result output data format
type Result struct {
	Code      int         `json:"code" example:"1"`
	Msg       string      `json:"message" example:"OK"`
	Data      interface{} `json:"data"`
	TimeStamp int64       `json:"timestamp" example:"1718850053"`
	TraceID   string      `json:"traceID" example:"7f0000016673920576f0659c620086b0"`
}

func newResp(c *gin.Context, code int, msg string, data interface{}) *Result {
	resp := &Result{
		Code:      code,
		Msg:       msg,
		TimeStamp: time.Now().Unix(),
		TraceID:   xctx.GetGinCtxTid(c),
	}

	// ensure that the data field is not nil on return, note that it is not nil when resp.data=[]interface {}, it is serialized to null
	if data == nil {
		resp.Data = &struct{}{}
	} else {
		resp.Data = data
	}

	// resp.Data = data

	return resp
}

var jsonContentType = []string{"application/json; charset=utf-8"}

func writeContentType(w http.ResponseWriter, value []string) {
	header := w.Header()
	if val := header["Content-Type"]; len(val) == 0 {
		header["Content-Type"] = value
	}
}

func writeJSON(c *gin.Context, code int, res interface{}) {
	c.Writer.WriteHeader(code)
	writeContentType(c.Writer, jsonContentType)
	err := json.NewEncoder(c.Writer).Encode(res)
	if err != nil {
		fmt.Printf("json encode error, err = %s\n", err.Error())
	}
}

func respJSONWithStatusCode(c *gin.Context, code int, msg string, data ...interface{}) {
	var firstData interface{}
	if len(data) > 0 {
		firstData = data[0]
	}
	resp := newResp(c, code, msg, firstData)

	writeJSON(c, code, resp)
}

// Output return standard HTTP status codes and message, parameter code is HTTP status code
func Output(c *gin.Context, code int, data ...interface{}) {
	switch code {
	case http.StatusOK:
		respJSONWithStatusCode(c, http.StatusOK, errcode.Success.Msg(), data...)
	case http.StatusBadRequest:
		respJSONWithStatusCode(c, http.StatusBadRequest, errcode.InvalidParams.Msg(), data...)
	case http.StatusUnauthorized:
		respJSONWithStatusCode(c, http.StatusUnauthorized, errcode.Unauthorized.Msg(), data...)
	case http.StatusForbidden:
		respJSONWithStatusCode(c, http.StatusForbidden, errcode.Forbidden.Msg(), data...)
	case http.StatusNotFound:
		respJSONWithStatusCode(c, http.StatusNotFound, errcode.NotFound.Msg(), data...)
	case http.StatusRequestTimeout:
		respJSONWithStatusCode(c, http.StatusRequestTimeout, errcode.Timeout.Msg(), data...)
	case http.StatusConflict:
		respJSONWithStatusCode(c, http.StatusConflict, errcode.Conflict.Msg(), data...)
	case http.StatusInternalServerError:
		respJSONWithStatusCode(c, http.StatusInternalServerError, errcode.InternalServerError.Msg(), data...)
	case http.StatusTooManyRequests:
		respJSONWithStatusCode(c, http.StatusTooManyRequests, errcode.LimitExceed.Msg(), data...)
	case http.StatusServiceUnavailable:
		respJSONWithStatusCode(c, http.StatusServiceUnavailable, errcode.ServiceUnavailable.Msg(), data...)

	default:
		respJSONWithStatusCode(c, code, http.StatusText(code), data...)
	}
}

// Out returns the standard HTTP status code and message, parameter err is errcode.Error
func Out(c *gin.Context, err *errcode.Error, data ...interface{}) {
	code := err.ToHTTPCode()
	switch code {
	case http.StatusOK:
		respJSONWithStatusCode(c, http.StatusOK, errcode.Success.Msg(), data...)
	case http.StatusInternalServerError:
		respJSONWithStatusCode(c, http.StatusInternalServerError, err.Msg(), data...)
	case http.StatusBadRequest:
		respJSONWithStatusCode(c, http.StatusBadRequest, err.Msg(), data...)
	case http.StatusUnauthorized:
		respJSONWithStatusCode(c, http.StatusUnauthorized, err.Msg(), data...)
	case http.StatusForbidden:
		respJSONWithStatusCode(c, http.StatusForbidden, err.Msg(), data...)
	case http.StatusNotFound:
		respJSONWithStatusCode(c, http.StatusNotFound, err.Msg(), data...)
	case http.StatusRequestTimeout:
		respJSONWithStatusCode(c, http.StatusRequestTimeout, err.Msg(), data...)
	case http.StatusConflict:
		respJSONWithStatusCode(c, http.StatusConflict, err.Msg(), data...)
	case http.StatusTooManyRequests:
		respJSONWithStatusCode(c, http.StatusTooManyRequests, err.Msg(), data...)
	case http.StatusServiceUnavailable:
		respJSONWithStatusCode(c, http.StatusServiceUnavailable, err.Msg(), data...)

	default:
		respJSONWithStatusCode(c, http.StatusNotExtended, err.Msg(), data...)
	}
}

// status code flat 200, custom error codes in data.code
func respJSONWith200(c *gin.Context, code int, msg string, data ...interface{}) {
	if len(data) > 0 {
		writeJSON(c, http.StatusOK, newResp(c, code, msg, data[0]))
		return
	}
	resp := newResp(c, code, msg, nil)
	writeJSON(c, http.StatusOK, resp)
}

// Success return success
func Success(c *gin.Context, data ...interface{}) {
	respJSONWith200(c, errcode.Success.Code(), errcode.Success.Msg(), data...)
}

// SuccessWithPage return success
func SuccessWithPage[T any](c *gin.Context, list []*T, total int64) {
	respJSONWith200(c, errcode.Success.Code(), errcode.Success.Msg(), gin.H{"list": list, "total": total})
}

// SuccessWithList return success
func SuccessWithList[T any](c *gin.Context, list []*T) {
	respJSONWith200(c, errcode.Success.Code(), errcode.Success.Msg(), gin.H{"list": list})
}

// ErrorE return error
func ErrorE(c *gin.Context, err *errcode.Error, data ...interface{}) {
	respJSONWith200(c, err.Code(), err.Msg(), data...)
}

// Error 处理API调用失败的错误，返回错误信息。
// 参数：
//   - c: gin.Context，当前的上下文对象。
//   - originErr: error，原始错误对象。
//   - outErr: errcode.Error，要返回的错误对象。
//   - msg: string，错误消息。
//   - data: ...interface{}，可选的附加数据。
func Error(c *gin.Context, err error) {
	// 如果原始错误不为空，则记录错误日志。
	msg := ""
	if err != nil {
		msg = err.Error()
		if xerr, ok := err.(*xerror.Error); ok {
			if !errcode.IsSysDefinedError(xerr.Code()) {
				logger.Error("api failed.", logger.Err(err), ctxUtil.GinTraceIDField(c))
			}
		} else {
			logger.Error("api failed.", logger.Err(err), ctxUtil.GinTraceIDField(c))
		}
	}
	// 使用respJSONWith200函数返回错误信息。
	// BFF层约定 失败传0 成功传1
	// respJSONWith200(c, outErr.Code(), msg, data...)
	respJSONWith200(c, CustomErrorCode, msg)
}
